import base64
import json
from urllib.request import Request, urlopen
import datetime

from django.conf import settings

def decode(m):
    return m.decode('utf-8')

def encode(m):
    return m.encode('utf-8')


class ConnectionInfo():
    """
    A class that contains information required to establish connections to the api
    """
    ACCOUNT_ID = settings.B2_ACCOUNT_ID
    APP_KEY = settings.B2_APPLICATION_KEY
    BUCKET_NAME = settings.B2_BUCKET_NAME
    BUCKET_ID = settings.B2_BUCKET_ID
    _validDuration = datetime.timedelta(days=1) # duration of the authorization token


    def __init__(self):
        self._auth_request_time = datetime.datetime.now() - 42*self._validDuration
        self._auth_data = dict()
        self.name_id_dict = dict() # To make get_file_id faster at the cost of memory


    @property
    def auth_data(self):
        """
        Returns the authorization data after sending a request with ID and APP_KEY. Updates the authorization token if necessary
        """
        # Check if the token is still valid
        if datetime.datetime.now() - self._auth_request_time >= self._validDuration:
            id_and_key = self.ACCOUNT_ID + ':' + self.APP_KEY
            basic_auth_string = 'Basic ' + decode(base64.b64encode(id_and_key.encode('utf-8')))
            headers = {'Authorization' : basic_auth_string}
            request = Request(
                'https://api.backblaze.com/b2api/v1/b2_authorize_account',
                headers = headers
            )
            response = urlopen(request)
            self._auth_request_time = datetime.datetime.now() # reset time
            self._auth_data = json.loads(decode(response.read()))
            response.close()
        return self._auth_data

    @property
    def auth_token(self):
        return self.auth_data['authorizationToken']

    @property
    def api_url(self):
        return self.auth_data['apiUrl']

    @property
    def download_url(self):
        return self.auth_data['downloadUrl']

    @property
    def minimumPartSize(self):
        return self.auth_data['minimumPartSize']

    @property
    def upload_data(self):
        """
        Returns information needed for uploads given BUCKET_ID and authorization data.
        Since every upload can potentially fail, we need to resend Request every time.
        """
        request = Request(
            '%s/b2api/v1/b2_get_upload_url' % self.api_url,
            encode(json.dumps({ 'bucketId' : self.BUCKET_ID })),
            headers = { 'Authorization': self.auth_token }
        )
        response = urlopen(request)
        _upload_data = json.loads(decode(response.read()))
        response.close()
        return _upload_data

    def get_file_id(self, name):
        """
        Returns the id of the file.
        Looking for way to do it more efficiently(without downloading the file!)
        """
        return self.name_id_dict[name]



import json
import mimetypes
import hashlib
from urllib.request import Request, urlopen

from django.core.files.base import ContentFile
from django.core.files.storage import Storage
from django.utils.deconstruct import deconstructible

from .connectioninfo import ConnectionInfo, decode, encode


@deconstructible
class B2Storage(Storage):
    def __init__(self, *args, **kwargs):
        super(B2Storage, self).__init__(*args, **kwargs)
        self.connection = ConnectionInfo()


    def _open(self, name, mode="rb"):
        url = self.connection.download_url + "/file/" + self.connection.BUCKET_NAME + "/" + name
        headers = {
            'Authorization' : self.connection.auth_token
        }
        request = Request(url, None, headers)
        response = urlopen(request)
        response_data = json.loads(decode(response.read()))
        response.close()
        return ContentFile(response_data)

    def _save(self, name, content):
        upload_data = self.connection.upload_data
        file_data = content.read()
        file_sha1 = hashlib.sha1(file_data).hexdigest()
        content_type = ""
        if hasattr(content.file, 'content_type'):
            content_type = content.file.content_type
        else:
            content_type = mimetypes.guess_type(name)[0]
        headers = {
            'Authorization' : upload_data['authorizationToken'],
            'X-Bz-File-Name' :  name,
            'Content-Type' : content_type,
            'X-Bz-Content-Sha1' : file_sha1
        }
        request = Request(upload_data['uploadUrl'], file_data, headers)
        response = urlopen(request)
        response_data = json.loads(decode(response.read()))
        # Add file_name:file_id to our dict
        file_id = response_data['fileId']
        name = response_data['fileName']
        self.connection.name_id_dict[name] = file_id
        response.close()
        return name


    def delete(self, name):
        file_id = self.connection.get_file_id(name)
        request = Request(
            '%s/b2api/v1/b2_delete_file_version' % self.connection.api_url,
            encode(json.dumps({ 'fileName' : name, 'fileId' : file_id })),
            headers = { 'Authorization': self.connection.auth_token }
        )
        response = urlopen(request)
        json.loads(decode(response.read()))
        response.close()
        del self.connection.name_id_dict[name]

    def url(self, name):
        return self.connection.download_url + '/file/' + self.connection.BUCKET_NAME + '/' + name


    def exists(self, name):
        return name in self.connection.name_id_dict



from setuptools import setup

def readme():
    with open('README.rst') as f:
        return f.read()

setup(
    name="django_b2storage",
    version='1.0',
    description='A backblaze b2 storage system for django',
    long_description=readme(),
    keywords='backblaze b2 django storage media',
    classifiers=[
        'Framework :: Django',
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Intended Audience :: Developers',
    ],
    url='https://github.com/amlatyrngom/django-b2storage',
    author='Amadou Latyr Ngom',
    author_email='amlatyr.ngom@gmail.com',
    licence='MIT',
    packages=['django_b2storage'],
    install_requires=[
        'Django >= 1.9',
    ],
    test_suite='nose.collector',
    tests_require=['nose'],
    include_package_data=True,
    zip_safe=False
)
